/**
 * Lodash Transfer From https://www.lodashjs.com/
 * @packageDocumentation
 */

/**
 * Lodash 是一个一致性、模块化、高性能的 JavaScript 实用工具库。
 * 
 * @document http://ser.yinengyun.com:8181/docs/pi_utils/pi_utils-1cm8armpps71p
 */
export namespace Lodash {

    /**@hidden */
    declare var global;

    /// https://github.com/lodash/lodash/blob/master/.internal/freeGlobal.js
    /** Detect free variable `global` from Node.js. */
    /**@hidden */
    export const freeGlobal = typeof global === 'object' && global !== null && global.Object === Object && global

    /// https://github.com/lodash/lodash/blob/master/.internal/root.js
    /** Detect free variable `globalThis` */
    /**@hidden */
    const freeGlobalThis = typeof globalThis === 'object' && globalThis !== null && globalThis.Object == Object && globalThis
    
    /** Detect free variable `self`. */
    /**@hidden */
    const freeSelf = typeof self === 'object' && self !== null && self.Object === Object && self
    
    /** Used as a reference to the global object. */
    /**@hidden */
    export const root = freeGlobalThis || freeGlobal || freeSelf || Function('return this')()

}

export namespace Lodash {
    /**
     * 延迟 `wait` 毫秒后调用 `func`。 调用时，任何附加的参数会传给 `func` 。
     *
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to delay.
     * @param {number} wait The number of milliseconds to delay invocation.
     * @param {...*} [args] The arguments to invoke `func` with.
     * @returns {number} Returns the timer id.
     * @example
     *
     * ```
     * delay(text => console.log(text), 1000, 'later')
     * // => Logs 'later' after one second.
     * ```
     */
    export function delay(func: Function, wait: number, ...args): number {
        if (typeof func !== 'function') {
        throw new TypeError('Expected a function')
        }
        return setTimeout(func, +wait || 0, ...args)
    }
}

export namespace Lodash {
    /**
     * `before` 的反向函数; 此方法创建一个函数，当他被调用 `n` 或更多次之后 `func` 才被真正执行
     * 
     * @since 0.1.0
     * @category Function
     * @param {number} n The number of calls before `func` is invoked.
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * ```
     * const saves = ['profile', 'settings']
     * const done = after(saves.length, () => console.log('done saving!'))
     *
     * forEach(saves, type => asyncSave({ 'type': type, 'complete': done }))
     * // => Logs 'done saving!' after the two async saves have completed.
     * ```
     */
    export function after(n: number, func: Function): Function {
        if (typeof func !== 'function') {
        throw new TypeError('Expected a function')
        }
        n = n || 0
        return function(...args) {
            if (--n < 1) {
                return func.apply(this, args)
            }
        }
    }
}

export namespace Lodash {
    /**
     * @language zh-cn https://www.lodashjs.com/docs/lodash.before
     * 
     * 创建一个调用 `func` 的函数，通过 `this` 绑定和创建函数的参数调用 `func` ，调用次数不超过 `n` 次。 之后再调用这个函数，将返回一次最后调用 `func` 的结果。
     *
     * @since 3.0.0
     * @category Function
     * @param {number} n The number of calls at which `func` is no longer invoked.
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * ```
     * jQuery(element).on('click', before(5, addContactToList))
     * // => Allows adding up to 4 contacts to the list.
     * ```
     */
    export function before(n: number, func: Function): Function {
        let result
        if (typeof func !== 'function') {
            throw new TypeError('Expected a function')
        }
        return function(...args) {
            if (--n > 0) {
                result = func.apply(this, args)
            }
            if (n <= 1) {
                func = undefined
            }
            return result
        }
    }
}

export namespace Lodash {
    /**
     * 
     * @language zh-cn https://www.lodashjs.com/docs/lodash.once
     * 
     * 创建一个只能调用 `func` 一次的函数。 
     * 重复调用返回第一次调用的结果。 
     * `func` 调用时， `this` 绑定到创建的函数，并传入对应参数。
     *
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new restricted function.
     * @example
     *
     * ```
     * const initialize = once(createApplication)
     * initialize()
     * initialize()
     * // => `createApplication` is invoked once
     * ```
     * @descript https://github.com/lodash/lodash/blob/master/once.js
     */
    export function once(func: Function): Function {
        return before(2, func);
    }
}

export namespace Lodash {
    /**
     * 检查 `value` 是否为 `Object` 的 `language type`[类型]。 (例如： arrays, functions, objects, regexes,new Number(0), 以及 new String(''))
     * 
     * Checks if `value` is the
     * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
     * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
     *
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an object, else `false`.
     * @example
     *
     * ```
     * isObject({})
     * // => true
     *
     * isObject([1, 2, 3])
     * // => true
     *
     * isObject(Function)
     * // => true
     *
     * isObject(null)
     * // => false
     * ```
     */
    export function isObject(value: any): boolean {
        const type = typeof value
        return value != null && (type === 'object' || type === 'function')
    }
}

export namespace Lodash {
    export interface IOptionDebounce {
        /**
         * 指定是否在延迟开始前调用。
         */
        leading: boolean;
        /**
         * 设置 func 允许被延迟的最大值 - ms
         */
        maxWait: number;
        /**
         * 指定是否在延迟结束后调用。
         */
        trailing: boolean;
    }
    export interface Debounce extends Function {
        /**
         * Cancel the trailing debounced invocation.
         */
        cancel: () => void;
        flush: () => void;
        /**
         * Check for pending invocations.
         */
        pending: () => void;
    }
    /**
     * 创建一个 debounced（防抖动）函数，该函数会从上一次被调用后，延迟 wait 毫秒后调用 func 方法。 
     * 
     * debounced（防抖动）函数提供一个 cancel 方法取消延迟的函数调用以及 flush 方法立即调用。 
     * 
     * 可以提供一个 options（选项） 对象决定如何调用 func 方法，options.leading 与|或 options.trailing 决定延迟前后如何触发（注：是 先调用后等待 还是 先等待后调用）。 
     * 
     * func 调用时会传入最后一次提供给 debounced（防抖动）函数 的参数。 
     * 
     * 后续调用的 debounced（防抖动）函数返回是最后一次 func 调用的结果。
     *
     * 
     * @note 如果 leading 和 trailing 选项为 true, 则 func 允许 trailing 方式调用的条件为: 在 wait 期间多次调用防抖方法。
     *
     * @note 如果 wait 为 0 并且 leading 为 false, func调用将被推迟到下一个点，类似setTimeout为0的超时。
     *
     * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
     * for details over the differences between `debounce` and `throttle`.
     *
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to debounce.
     * @param {number} [wait=0]
     *  The number of milliseconds to delay; if omitted, `requestAnimationFrame` is
     *  used (if available).
     * @param {Object} [options={}] The options object.
     * @param {boolean} [options.leading=false]
     *  Specify invoking on the leading edge of the timeout.
     * @param {number} [options.maxWait]
     *  The maximum time `func` is allowed to be delayed before it's invoked.
     * @param {boolean} [options.trailing=true]
     *  Specify invoking on the trailing edge of the timeout.
     * @returns {Function} Returns the new debounced function.
     * @example
     *
     * ```
     * // Avoid costly calculations while the window size is in flux.
     * jQuery(window).on('resize', debounce(calculateLayout, 150))
     *
     * // Invoke `sendMail` when clicked, debouncing subsequent calls.
     * jQuery(element).on('click', debounce(sendMail, 300, {
     *   'leading': true,
     *   'trailing': false
     * }))
     *
     * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
     * const debounced = debounce(batchLog, 250, { 'maxWait': 1000 })
     * const source = new EventSource('/stream')
     * jQuery(source).on('message', debounced)
     *
     * // Cancel the trailing debounced invocation.
     * jQuery(window).on('popstate', debounced.cancel)
     *
     * // Check for pending invocations.
     * const status = debounced.pending() ? "Pending..." : "Ready"
     * ```
     */
    export function debounce(func: Function, wait: number, options: IOptionDebounce): Debounce {
        let lastArgs,
        lastThis,
        maxWait,
        result,
        timerId,
        lastCallTime
    
        let lastInvokeTime = 0
        let leading = false
        let maxing = false
        let trailing = true
    
        // Bypass `requestAnimationFrame` by explicitly setting `wait=0`.
        const useRAF = (!wait && wait !== 0 && typeof root.requestAnimationFrame === 'function')
    
        if (typeof func !== 'function') {
            throw new TypeError('Expected a function')
        }
        wait = +wait || 0
        if (isObject(options)) {
            leading = !!options.leading
            maxing = 'maxWait' in options
            maxWait = maxing ? Math.max(+options.maxWait || 0, wait) : maxWait
            trailing = 'trailing' in options ? !!options.trailing : trailing
        }
    
        function invokeFunc(time) {
            const args = lastArgs
            const thisArg = lastThis
        
            lastArgs = lastThis = undefined
            lastInvokeTime = time
            result = func.apply(thisArg, args)
            return result
        }
    
        function startTimer(pendingFunc, wait) {
            if (useRAF) {
                root.cancelAnimationFrame(timerId)
                return root.requestAnimationFrame(pendingFunc)
            }
            return setTimeout(pendingFunc, wait)
        }

        function cancelTimer(id) {
            if (useRAF) {
                return root.cancelAnimationFrame(id)
            }
            clearTimeout(id)
        }

        function leadingEdge(time) {
            // Reset any `maxWait` timer.
            lastInvokeTime = time
            // Start the timer for the trailing edge.
            timerId = startTimer(timerExpired, wait)
            // Invoke the leading edge.
            return leading ? invokeFunc(time) : result
        }

        function remainingWait(time) {
            const timeSinceLastCall = time - lastCallTime
            const timeSinceLastInvoke = time - lastInvokeTime
            const timeWaiting = wait - timeSinceLastCall
        
            return maxing
                ? Math.min(timeWaiting, maxWait - timeSinceLastInvoke)
                : timeWaiting
        }

        function shouldInvoke(time) {
            const timeSinceLastCall = time - lastCallTime
            const timeSinceLastInvoke = time - lastInvokeTime
        
            // Either this is the first call, activity has stopped and we're at the
            // trailing edge, the system time has gone backwards and we're treating
            // it as the trailing edge, or we've hit the `maxWait` limit.
            return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
                (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait))
        }

        function timerExpired() {
            const time = Date.now()
            if (shouldInvoke(time)) {
                return trailingEdge(time)
            }
            // Restart the timer.
            timerId = startTimer(timerExpired, remainingWait(time))
        }

        function trailingEdge(time) {
            timerId = undefined
    
            // Only invoke if we have `lastArgs` which means `func` has been
            // debounced at least once.
            if (trailing && lastArgs) {
                return invokeFunc(time)
            }
            lastArgs = lastThis = undefined
            return result
        }
    
        function cancel() {
            if (timerId !== undefined) {
                cancelTimer(timerId)
            }
            lastInvokeTime = 0
            lastArgs = lastCallTime = lastThis = timerId = undefined
        }
    
        function flush() {
            return timerId === undefined ? result : trailingEdge(Date.now())
        }
    
        function pending() {
            return timerId !== undefined
        }
    
        function debounced(...args) {
            const time = Date.now()
            const isInvoking = shouldInvoke(time)
        
            lastArgs = args
            lastThis = this
            lastCallTime = time
        
            if (isInvoking) {
                if (timerId === undefined) {
                    return leadingEdge(lastCallTime)
                }
                if (maxing) {
                    // Handle invocations in a tight loop.
                    timerId = startTimer(timerExpired, wait)
                    return invokeFunc(lastCallTime)
                }
            }
            if (timerId === undefined) {
                timerId = startTimer(timerExpired, wait)
            }
            return result
        }
        debounced.cancel = cancel
        debounced.flush = flush
        debounced.pending = pending
        return debounced
    }
    
}

export namespace Lodash {
    /**
     * 创建一个针对断言函数 `func` 结果取反的函数。 `func` 断言函数被调用的时候， `this` 绑定到创建的函数，并传入对应参数。
     *
     * @since 3.0.0
     * @category Function
     * @param {Function} predicate The predicate to negate.
     * @returns {Function} Returns the new negated function.
     * @example
     *
     * ```
     * function isEven(n) {
     *   return n % 2 == 0
     * }
     *
     * filter([1, 2, 3, 4, 5, 6], negate(isEven))
     * // => [1, 3, 5]
     * ```
     */
    export function negate(predicate: Function): Function {
        if (typeof predicate !== 'function') {
            throw new TypeError('Expected a function')
        }
        return function(...args) {
            return !predicate.apply(this, args)
        }
    }
}

export namespace Lodash {
        
    /**
     * 创建一个节流函数，在 `wait` 毫秒内最多执行 `func` 一次的函数。 
     * 
     * 该函数提供一个 `cancel` 方法取消延迟的函数调用以及 `flush` 方法立即调用。 
     * 
     * 可以提供一个 `options` 对象决定如何调用 func 方法， `options.leading` 与|或 `options.trailing `决定 `wait` 前后如何触发。 
     * 
     * `func` 会传入最后一次传入的参数给这个函数。 
     * 
     * 随后调用的函数返回是最后一次 `func` 调用的结果。
     *
     * @note 如果 `leading` 和 `trailing` 都设定为 `true` 则 `func` 允许 `trailing` 方式调用的条件为: 在 `wait` 期间多次调用。
     * 
     * @note 如果 `wait` 为 `0` 并且 `leading` 为 `false`, `func` 调用将被推迟到下一个点，类似 `setTimeout` 为 `0` 的超时
     *
     * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
     * for details over the differences between `throttle` and `debounce`.
     *
     * @since 0.1.0
     * @category Function
     * @param {Function} func The function to throttle.
     * @param {number} [wait=0]
     *  The number of milliseconds to throttle invocations to; if omitted,
     *  `requestAnimationFrame` is used (if available).
     * @param {Object} [options={}] The options object.
     * @param {boolean} [options.leading=true]
     *  Specify invoking on the leading edge of the timeout.
     * @param {boolean} [options.trailing=true]
     *  Specify invoking on the trailing edge of the timeout.
     * @returns {Function} Returns the new throttled function.
     * @example
     *
     * ```
     * // Avoid excessively updating the position while scrolling.
     * jQuery(window).on('scroll', throttle(updatePosition, 100))
     *
     * // Invoke `renewToken` when the click event is fired, but not more than once every 5 minutes.
     * const throttled = throttle(renewToken, 300000, { 'trailing': false })
     * jQuery(element).on('click', throttled)
     *
     * // Cancel the trailing throttled invocation.
     * jQuery(window).on('popstate', throttled.cancel)
     * ```
     */
    export function throttle(func: Function, wait: number, options: IOptionDebounce): Function {
        let leading = true
        let trailing = true
    
        if (typeof func !== 'function') {
            throw new TypeError('Expected a function')
        }
        if (isObject(options)) {
            leading = 'leading' in options ? !!options.leading : leading
            trailing = 'trailing' in options ? !!options.trailing : trailing
        }
        return debounce(func, wait, {
            leading,
            trailing,
            'maxWait': wait
        })
    }
}