// ============================== 导入
import { imul, INT32_SIZE, INT32_MAX, UINT32_MAX, add64, mul64 } from "./math"

// ============================== 导出

const BIT_53 = 9007199254740992.0;
const BIT_27 = 134217728.0;

export class Rand {
	private data: Random;

	constructor(seed = 0, type = 0) {
		this.data = (type === 0) ? new Pcg(seed) : (type === 1) ? new LC(seed) : new RC4(seed);
	}
	public init(data: Random) {
		this.data = data;
	}
	public next(): number {
		return this.data.next();
	}
	public discard(count: number) {
		this.data.discard(count);
	}
	public getState() {
		return this.data.getState();
	}
	public setState(state: any) {
		this.data.setState(state);
	}
	/// Get a uniformly distributed IEEE-754 double between 0.0 and 1.0, with
	/// 53 bits of precision (every bit of the mantissa is randomized).
	public nextF64(): number {
		const hi = (this.next() & 0x03ffffff) * 1.0;
		const lo = (this.next() & 0x07ffffff) * 1.0;
		return ((hi * BIT_27) + lo) / BIT_53;
	}
	/**
	 * @description 获得一个指定范围（左闭右开区间）的随机浮点数
	 */
	public nextFloat(v1: number, v2: number): number {
		return v1 + this.next() / UINT32_MAX * (v2 - v1);
	}
	/**
	 * @description 获得一个指定范围（闭区间）的随机整数
	 */
	public nextInt(v1: number, v2: number): number {
		return v1 + Math.floor(this.next() / UINT32_MAX * (v2 - v1 + 1.0));
	}
	/**
	 * @description 数组乱序
	 */
	public shuffle(array: any[]) {
		for (let i = array.length - 1; i >= 0; i--) {
			const j = Math.floor(this.next() / UINT32_MAX * (i + 1));
			const temp = array[i];
			array[i] = array[j];
			array[j] = temp;
		}
	}
}

export interface Random {
	next(): number;
	discard(count: number);
	getState(): any;
	setState(state: any);
}

const INC_HI = 0x14057b7e;
const INC_LO = 0xf767814f;
const MUL_HI = 0x5851f42d;
const MUL_LO = 0x4c957f2d;

/**
 * PCG 的随机种子. 提供32位随机性
 * Generate a random 32 bit integer. This uses the PCG algorithm, described here: http://www.pcg-random.org/
* https://github.com/thomcc/pcg-random
 */
export class Pcg implements Random {
	state: Uint32Array;
	constructor(seedHi = 0, seedLo = 0, incHi = 0, incLo = 0) {
		if (!seedLo) {
			seedLo = seedHi;
			seedHi = 0;
		}
		seedLo = seedLo || (Math.random() * 0xffffffff) >>> 0;
		if (!incLo) {
			if (!incHi) {
				incLo = INC_LO;
				incHi = INC_HI;
			} else {
				incLo = incHi;
				incHi = 0;
			}
		}
		this.state = new Uint32Array([seedHi >>> 0, seedLo >>> 0, incHi >>> 0, incLo >>> 0]);
		this.discard(1);
	}
	next(): number {
		// save current state (what we'll use for this number)
		const oldHi = this.state[0];
		const oldLo = this.state[1];
		// churn LCG.
		mul64(oldHi, oldLo, MUL_HI, MUL_LO, this.state);
		add64(this.state[0], this.state[1], this.state[2], this.state[3], this.state);

		// get least sig. 32 bits of ((oldstate >> 18) ^ oldstate) >> 27
		let xsHi = oldHi >>> 18;
		let xsLo = ((oldLo >>> 18) | (oldHi << 14)) >>> 0;
		xsHi = (xsHi ^ oldHi) >>> 0;
		xsLo = (xsLo ^ oldLo) >>> 0;
		const xorshifted = ((xsLo >>> 27) | (xsHi << 5)) >>> 0;
		// rotate xorshifted right a random amount, based on the most sig. 5 bits
		// bits of the old state.
		const rot = oldHi >>> 27;
		const rot2 = ((-rot >>> 0) & 31) >>> 0;
		return ((xorshifted >>> rot) | (xorshifted << rot2)) >>> 0;
	}
	discard(count: number) {
		while (count-- > 0) {
			mul64(this.state[0], this.state[1], MUL_HI, MUL_LO, this.state);
			add64(this.state[0], this.state[1], this.state[2], this.state[3], this.state);
		}
	}
	/// Get the state of the random number generator.
	getState() {
		return this.state.slice()
	}
	/// Set the state of the random number generator.
	setState(state: Uint32Array) {
		this.state = state;
	}
}

const ARRAY_SIZE = 624;
const ARRAY_MAX = ARRAY_SIZE - 1;
const M = 397;
const ARRAY_SIZE_MINUS_M = ARRAY_SIZE - M;
const A = 0x9908b0df;

/**
 * RC4 的随机种子. 提供32位随机性
 * https://github.com/ckknight/random-js
 */
export class RC4 implements Random {

	private data: Int32Array;
	private index: number; // integer within [0, 624]

	constructor(seed = 0) {
		let data = new Int32Array(ARRAY_SIZE);
		let previous = 0;
		data[0] = previous = (seed || (Math.random() * 0xffffffff)) >>> 0;

		for (let i = 1; i < ARRAY_SIZE; i++) {
			data[i] = previous = (imul(previous ^ (previous >>> 30), 0x6c078965) + i) | 0;
		}
		this.data = data;
		this.index = ARRAY_SIZE;
	}
	public next(): number {
		if (this.index >= ARRAY_SIZE) {
			this.refreshData(this.data);
			this.index = 0;
		}
		const v = this.data[this.index];
		this.index++;
		return this.value(v);
	}
	public discard(count: number) {
		if (count <= 0) {
			return;
		}
		if (this.index >= ARRAY_SIZE) {
			this.refreshData(this.data);
			this.index = 0;
		}
		while (count + this.index > ARRAY_SIZE) {
			count -= ARRAY_SIZE - this.index;
			this.refreshData(this.data);
			this.index = 0;
		}
		this.index += count;
	}
	public getState() {
		return { data: this.data, index: this.index }
	}
	public setState(state: { data: Int32Array, index: number }) {
		this.data = state.data;
		this.index = state.index;
	}
	private value(v: number) {
		v ^= v >>> 11;
		v ^= (v << 7) & 0x9d2c5680;
		v ^= (v << 15) & 0xefc60000;
		return (v ^ (v >>> 18)) >>> 0;
	}
	private refreshData(data: Int32Array) {
		let k = 0;
		let tmp = 0;
		for (; k < ARRAY_SIZE_MINUS_M; k++) {
			tmp = (data[k] & INT32_SIZE) | (data[(k + 1)] & INT32_MAX);
			data[k] = data[(k + M)] ^ (tmp >>> 1) ^ (tmp & 0x1 ? A : 0);
		}
		for (; k < ARRAY_MAX; k++) {
			tmp = (data[k] & INT32_SIZE) | (data[(k + 1)] & INT32_MAX);
			data[k] = data[(k - ARRAY_SIZE_MINUS_M)] ^ (tmp >>> 1) ^ (tmp & 0x1 ? A : 0);
		}
		tmp = (data[ARRAY_MAX] & INT32_SIZE) | (data[0] & INT32_MAX);
		data[ARRAY_MAX] = data[M - 1] ^ (tmp >>> 1) ^ (tmp & 0x1 ? A : 0);
	}
}

const RANDOM_A = 16807;
const RANDOM_Q = 127773;
const RANDOM_MASK = 123459876;
const RANDOM_AQR = INT32_MAX;

// Linear congruence 线性同余算法
export class LC implements Random {
	private seed: number;
	constructor(seed = 0) {
		this.seed = (seed || (Math.random() * 0xffffffff)) >>> 0;
	}
	/**
	 * 返回值 [0, MAX_INT32]
	 */
	public next(): number {
		this.seed ^= RANDOM_MASK;
		let k = Math.floor(this.seed / RANDOM_Q);
		this.seed = RANDOM_A * this.seed - k * RANDOM_AQR;
		if (this.seed < 0) {
			this.seed += INT32_MAX;
		}
		return this.seed;
	}
	public discard(count: number) {
		if (count <= 0) {
			return;
		}
		while (count--) {
			this.next();
		}
	}
	public getState() {
		return this.seed;
	};
	public setState(state: number) {
		this.seed = state;
	}
}
