/**
 * ```
 * 提供 支持的资源类型标识
 * 提供  resTab.load 方法加载对应资源时的 类型表达
 * 封装 PiTexture 作为纹理资源
 * ```
 * @packageDocumentation
 */

import { isPowerOfTwo } from '../math/math';
import { PISYS } from 'pi_sys/index';
import { Res } from './res';
import { register, ResTab } from "./res_mgr";
import { IParamPiFont } from 'pi_sys/render/font';

/**
 * @hidden
 */
export function updateTextureCreate() {

}

/**
 * 资源加载模块初始化
 * 注册常用资源类型的加载流程
 * @mark 已加入 pi_uitls/main.ts > init_pi_utils() , 作为 pi_utils 初始化的一部分
 */
export function init() {
    register<typeof RES_IMAGE, any, PISYS.PiImage>(
        RES_IMAGE, 
        (resTab: ResTab, type: string = RES_IMAGE, filePath: string, ...args: any[]) => {
            return new Promise((resolve: (res: PISYS.PiImage) => void, reject) => {
                PISYS.ResLoadTool.loadImage(filePath, args)
                .then(res => {
                    resolve(res);
                })
                .catch((e) => {
                    console.warn(e);

                    const image = PISYS.createImage();

                    image.onload = () => {
                        resolve(image);
                    }
                    image.onerror = (e) => {
                        reject(e);
                    }
                    image.src = filePath;
                });
            });
            // return PISYS.ResLoadTool.loadImage(filePath, args);
        },
        (link: PISYS.PiImage) => {
            link.src = '';
        }
    );
    
    register<typeof RES_FONT, any, PISYS.PiFont>(
        RES_FONT, 
        (resTab: ResTab, type: string = RES_FONT, filePath: string, ...args: [IParamPiFont]) => {
            return PISYS.ResLoadTool.loadFont(filePath, args);
        },
        (link: PISYS.PiFont) => {
            link.src = '';
        }
    );
    
    register<typeof RES_AUDIO, any, PISYS.PiAudio>(
        RES_AUDIO, 
        (resTab: ResTab, type: string = RES_AUDIO, filePath: string, ...args: any[]) => {
            return PISYS.ResLoadTool.loadAudio(filePath, args);
        },
        (link: PISYS.PiAudio) => {
            link.destroy();
        }
    );
    
    register<typeof RES_FILE, any, ArrayBuffer | Uint8Array>(
        RES_FILE, 
        (resTab: ResTab, type: string = RES_FILE, filePath: string, ...args: any[]) => {
            return PISYS.ResLoadTool.loadFile(filePath, args);
        },
        (link: ArrayBuffer) => {
            // link;
        }
    );
    
    register<typeof RES_PITEXTURE, IParamPiTexture, PiTexture>(
        RES_PITEXTURE, 
        (resTab: ResTab, type: typeof RES_PITEXTURE, filePath: string, ...args: [IParamPiTexture]): Promise<PiTexture> => {
            const piTex = new PiTexture();
            piTex.args = args;
            piTex.src = filePath;

            return new Promise((resolve, reject) => {
                piTex.onloaded = () => {
                    resolve(piTex);
                }
                piTex.onerror = (e) => {
                    reject(e);
                }
            });
        },
        (link: PiTexture) => {
            link.destroy();
        }
    );
}

/**
 * 统一的纹理资源
 */
export class PiTexture {
    /**
     * gl webgl 上下文
     * @mark 必须在应用初始化时设置 gl
     */
    public static gl: WebGLRenderingContext;
    /**
     * 纹理宽度 - 像素
     * @hidden
     */
    private _width: number = 0;
    /**
     * 纹理高度 - 像素
     * @hidden
     */
    private _height: number = 0;
    /**
     * gl纹理对象
     * @hidden
     */
    private _texture: WebGLTexture = null;

    /**
     * 纹理路径
     * @hidden
     */
    private _url: string = '';
    /**
     * 销毁标识
     * @hidden
     */
    private _destroyed: boolean = false;
    /**
     * 所属资源表
     * @hidden
     */
    private resTab: ResTab = new ResTab();
    /**
     * 资源文件路径
     */
    public set src(v: string) {
        this._url = v;
        this.load();
    }
    /**
     * 资源文件路径
     */
    public get src() {
        return this._url;
    }
    /**
     * 纹理宽度 - 像素
     */
    public get width() {
        return this._width;
    }
    /**
     * 纹理高度 - 像素
     */
    public get height() {
        return this._height;
    }
    /**
     * gl纹理对象
     */
    public get webglTex() {
        return this._texture;
    }
    constructor() {
    }
    /**
     * 创建时参数
     */
    public args: [IParamPiTexture?];
    /**
     * 纹理格式
     */
    public format: number;
    /**
     * 是否 noMipmap
     */
    public noMipmap: boolean;
    /**
     * 是否反转Y
     */
    public invertY: boolean;
    /**
     * 采样模式
     */
    public samplingMode: number;
    /**
     * 加载成功回调
     */
    public onloaded: (tex: PiTexture) => void;
    /**
     * 加载错误回调
     */
    public onerror: (e) => void;

    /**
     * 加载
     * @hidden
     */
    private load() {
        if (this._url !== '') {
            if (this._url.endsWith('.jpg')) {
                this.format = PiTexture.gl.RGB;
            }
            else {
                this.format = PiTexture.gl.RGBA;
            }
        }
        this.resTab.load(RES_IMAGE, this._url, []).then((res) => {
            const image = res.link;
            
            if (!this._destroyed) {
                this._loaded(image);

                this.onloaded && this.onloaded(this);
            }

            this.resTab.delete(res);
        })
        .catch((e) => {
            // this.onerror && this.onerror(e);
            console.warn(e);

            const image = new Image();
            
            image.onload = () => {
                if (!this._destroyed) {
                    this._loaded(image);
    
                    this.onloaded && this.onloaded(this);
                }
            }

            image.onerror = (e) => {
                this.onerror && this.onerror(e);
            }

            image.src = this._url;
        });
    }

    /**
     * 加载后处理
     * @tip IParamPiTexture 提供参考，决定目标纹理是否使用为 非2的幂尺寸内容
     * @tip 作为2的幂尺寸的纹理时，是否启用 mipMap
     * @tip 作为2的幂尺寸的纹理时 其他设置由使用者自己再处理
     * @param image 加载到的图片
     * @hidden
     */
    private _loaded(image: PISYS.PiImage) {
        const [param] = this.args;

        const gl = PiTexture.gl;

        if (!param || !param.mipMap || !param.useForPowerOfTwo) {
            this._width     = image.width;
            this._height    = image.height;
            this._texture   = gl.createTexture();

            gl.activeTexture(0);
            gl.bindTexture(gl.TEXTURE_2D, this._texture);
            gl.texImage2D(gl.TEXTURE_2D, 0, this.format, this.format, gl.UNSIGNED_BYTE, image);

            // No, it's not a power of 2. Turn of mips and set wrapping to clamp to edge
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);

            gl.bindTexture(gl.TEXTURE_2D, null);
        }
        else
        {
            const is2 = isPowerOfTwo(image.width) && isPowerOfTwo(image.height);

            this._width     = image.width;
            this._height    = image.height;
            this._texture   = PiTexture.gl.createTexture();

            gl.activeTexture(0);
            gl.bindTexture(gl.TEXTURE_2D, this._texture);

            if (!is2) {
                console.error(`图片尺寸需要为 2 的幂 - ${this._url} - ${image.src}`);

                // Fill the texture with a 1x1 blue pixel. - 洋红色
                gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, 1, 1, 0, gl.RGBA, gl.UNSIGNED_BYTE,
                    new Uint8Array([207, 0, 112, 255]));
            }
            else
            {
                gl.texImage2D(gl.TEXTURE_2D, 0, this.format, this.format, gl.UNSIGNED_BYTE, image);
            }

            if (param.mipMap) {
                // Yes, it's a power of 2. Generate mips.
                gl.generateMipmap(gl.TEXTURE_2D);
            }

            gl.bindTexture(gl.TEXTURE_2D, null);
        }
    }
    /**
     * 销毁
     */
    public destroy() {
        this._destroyed = true;
        if (this._texture) {
            PiTexture.gl.deleteTexture(this._texture);
            this._texture = null;
        }
        this.resTab.clear();
    }
}

/**
 * 纹理资源创建参数
 */
export interface IParamPiTexture {
    /**
     * 是否转换为 2 的幂尺寸内容
     */
    useForPowerOfTwo?: boolean;
    /**
     * 是否激活 mipMap
     */
    mipMap?: boolean;
}

/**
 * image 资源类型标识
 */
export const RES_IMAGE      = 'RES_IMAGE';
/**
 * 字体 资源类型标识
 */
export const RES_FONT       = 'RES_FONT';
/**
 * 音频 资源类型标识
 */
export const RES_AUDIO      = 'RES_AUDIO';
/**
 * 一般文件 资源类型标识
 */
export const RES_FILE       = 'RES_FILE';
/**
 * 纹理 资源类型标识
 */
export const RES_PITEXTURE  = 'RES_PITEXTURE';

/**
 * resTab.load - 加载 RES_PITEXTURE 时的类型表达
 */
export type F_RES_PITEXTURE   = (type: typeof RES_PITEXTURE, filePath: string, args: [IParamPiTexture?], hasTabRes?: boolean) => Promise<Res<IParamPiTexture, PiTexture>>;
/**
 * resTab.load - 加载 RES_IMAGE 时的类型表达
 */
export type F_RES_IMAGE       = (type: typeof RES_IMAGE, filePath: string, args: [any?], hasTabRes?: boolean) => Promise<Res<any, PISYS.PiImage>>;
/**
 * resTab.load - 加载 RES_FONT 时的类型表达
 */
export type F_RES_FONT        = (type: typeof RES_FONT, filePath: string, args: [any?], hasTabRes?: boolean) => Promise<Res<any, PISYS.PiFont>>;
/**
 * resTab.load - 加载 RES_AUDIO 时的类型表达
 */
export type F_RES_AUDIO       = (type: typeof RES_AUDIO, filePath: string, args: [any?], hasTabRes?: boolean) => Promise<Res<any, PISYS.PiAudio>>;
/**
 * resTab.load - 加载 RES_FILE 时的类型表达
 */
export type F_RES_FILE        = (type: typeof RES_FILE, filePath: string, args: [any?], hasTabRes?: boolean) => Promise<Res<any, ArrayBuffer | Uint8Array>>;