/**
 * 通用的资源管理，二级缓冲。
 * ```
 * 1、注册，以 Texture 为例，前提：image已经注册到res去了。
 *
 *   register("texture", (resTab, type, name, path, imageConfig) => {
 *          // load的最后一个参数，代表需不需要为该资源添加引用计数
 *          // 如果是外部人员使用，默认是true；
 *          // 这里用于图片仅仅是加载纹理的中间对象，不要缓冲，传false
 *          return resTab.load("image", name, [path, imageConfig], false).then(imageRes => {
 *              // 在这里调用创建纹理的实现
 *              let texture = createTexture(imageRes.link);
 *
 *              return Promise.resolve(texture);
 *           });
 *    }, (texture) => {
 *       // 在这里释放纹理，texture就是上面加载函数创建的纹理
 *       destroyTexture(texture);
 *    });
 *
 * 2、加载，以纹理为例：
 *
 *    resTab.load("texture", name, [path, config]).then(textureRes => {
 *       // 纹理就是 textureRes.link
 *    });
 * ```
 * @TODO TODO 存在 WASM asm 两个版本，构建需要处理
 * @packageDocumentation
 */

import { ResMgr, ResType, createResType } from "./res_mgr/mod";
import { PISYS } from "pi_sys/index";
import { genKey, Res as _Res } from "./res";

/**
 * 底层资源管理器对象
 * @hidden
 */
export const RES_MGR = new ResMgr(60 * 1024 * 1024);

/**
 * 资源引用
 * @hidden
 */
export interface ResRefLink {
    /**
     * 资源数据
     */
    data: any;
    /**
     * 
     */
    link()
}

/**
 * 资源管理表
 */
export class ResTab {
    /**
     * @hidden
     */
    private map: Map<string, _Res<ResArg, ResData>> = new Map();
    /**
     * @hidden
     */
    private resMap: Map<_Res<ResArg, ResData>, _Res<ResArg, ResData>> = new Map();
    /**
     * 
     * @param type 资源类型
     * @param name 资源名称/路径
     * @param loadArgs 加载参数
     * @param hasTabRes 弃用
     */
    public load = (type: any, name: string, loadArgs: ResArg[], hasTabRes = true): Promise<_Res<ResArg, ResData>> => {
        
        // window.logg = window.logg || {};
        // window.logg[name] = window.logg[name] || { count: 0 };
        // window.logg[name].count++;

        let key = genKey(type, name);

        let ref = RES_MGR.get_res(resTypeMap.get(type), key);

        let res = new Res<ResArg, ResData>();

        this.resMap.set(res, res);

        if (ref) {
            res.create(name, type, ref, loadArgs);
            return Promise.resolve(res);
        } else {
            
        }

        // // 取到就返回，内部做了引用+1了
        // let res = this.map.get(key);
        // if (res) {
        //     if (res.getContext()) {
        //         return Promise.resolve(res);
        //     }
        // }

        // 等待的Promise
        let wait = waitMap.get(key);
        if (!wait) {
            // 取对应类型的加载函数
            let func = createMap.get(type);
            if (!func) {
                throw new Error("res_mgr load failed, type isn't registered, type = " + type);
            }

            // 等待
            wait = func.load(this, type, name, ...loadArgs).then((data) => {
                // 加载完成，创建res，并从等待移除
                waitMap.delete(key);

                return data; // this.createRes(name, type, data, loadArgs);
            }).catch((err) => {
                // 加载失败，从等待移除，从新抛出错误。
                // 注：暂时不能用Promise.finally，因为需要浏览器版本比较新，Chrome 63
                waitMap.delete(key);
                
                this.resMap.delete(res);
                return Promise.reject(err);
            });

            // 将Promise设置到等待列表
            waitMap.set(key, wait);
        }

        // 基于wait处理tab的信息
        let p = wait.then(
            // (res) => {
            (data) => {
                // 异步回来的时候，可能tab已经被release
                if (!this.map) {
                    res.free();
                    return Promise.reject(new Error("res_mgr load failed, resTab has released"));
                }

                // // if (hasTabRes) {
                // //     this.addRes(key, res);
                // // } else if (res) {
                // //     // // 如果当前的res不被任何tab引用，则开始释放流程
                // //     // const t = PISYS.Time.now();
                // //     // timeoutRelease(res, t, t + defalutTimeout);
                // //     res.free();
                // // }
                // return res;

                const sizeOf = createMap.get(type).sizeOf;
                const key = genKey(type, name);
                const size = sizeOf ? sizeOf(data) : PISYS.Depend.getFile(name).size;

                let ref = RES_MGR.get_res(resTypeMap.get(type), key);
                if (!ref) {
                    ref = RES_MGR.create_res(
                        resTypeMap.get(type), 
                        key,
                        {
                            data: data,
                            destroy: () => {
                                const func = createMap.get(type).destroy;
                                func && func(data);
                                // console.log(res.key);
                            }
                        }, 
                        size
                    );
                }
                res.create(name, type, ref, loadArgs);
                return res;
            }
        )
        .catch(
            (e) => {
                this.resMap.delete(res);
                return Promise.reject(new Error(`res_mgr load failed, ${e}`));
            }
        ) ;

        return p;
    }
    
    /**
     * 提供给逻辑内存资源，同步创建Res
     * @param name 资源名称/路径
     * @param type 资源类型
     * @param data 资源数据
     * @param loadArgs 加载参数
     */
    public createRes(name: string, type: string, data: any, loadArgs: any): _Res<ResArg, ResData> {
        // window.logg = window.logg || {};
        // window.logg[name] = window.logg[name] || { count: 0 };
        // window.logg[name].count++;

        let r = new Res<ResArg, ResData>();
        const sizeOf = createMap.get(type).sizeOf;
        const key = genKey(type, name);
        const size = sizeOf ? sizeOf(data) : PISYS.Depend.getFile(name).size;

        let ref = RES_MGR.get_res(resTypeMap.get(type), key);
        if (!ref) {
            ref = RES_MGR.create_res(
                resTypeMap.get(type), 
                key,
                {
                    data: data,
                    destroy: () => {
                        const func = createMap.get(type).destroy;
                        func && func(data);
                        console.log(r.key);
                    }
                }, 
                size
            );
        }
        r.create(name, type, ref, loadArgs);
        this.resMap.set(r, r);
        return r;
    }

    /**
     * @description 释放资源
     * @param res 资源引用
     * @returns 是否删除成功
     * @example
     */
    public delete(res: _Res<ResArg, ResData>): boolean {
        this.resMap.delete(res);
        res.free();
        // window.logg[res.name].count--;
        return true;

        // const tab = this.map;
        
        // if (!tab || !res) { // TODO 可以不判断res是否存在， 临时解决pibabylon的资源释放问题
        //     return false;
        // }
        // const b = tab.delete(res.key);
        // if (b) {
        //     res.free();
        // }

        // return b;
    }

    /**
     * @description 清除全部资源
     * @example
     */
    public clear(): void {
        for (const res of this.map.values()) {
            res.free();
        }
        this.map.clear();
        
        for (const res of this.resMap.values()) {
            res.free();
        }
        this.resMap.clear();
    }
    /**
     * 获取资源引用
     * @param type 资源类型
     * @param name 资源名称/路径
     */
    public get(type: string, name: string) {
        let key = genKey(type, name);

        let ref = RES_MGR.get_res(resTypeMap.get(type), key);

        if (ref) {
            let r = new Res<ResArg, ResData>();
            this.resMap.set(r, r);
            r.create(name, type, ref, []);
            return r;
        } else {
            return undefined;
        }
    }
}

/**
 * @hidden
 */
export type TLoadCall<T, A, D> = (resTab: ResTab, type: T, filePath: string, ...args: A[]) => Promise<D>;

/**
 * @hidden
 */
type ResArg = any;
/**
 * @hidden
 */
type ResData = any;

/**
 * 注册：类型名，以及该类型的加载函数，还有销毁函数
 * @param load 加载函数，返回Promise<link>，通过返回值，可以取到 将要 放到res.link的数据
 * @param destroy 销毁函数，用于销毁res.link
 * @param sizeOf 查询资源的字节大小方法
 * @param replace load 方法是否明确为重载
 */
export function register<T, A, D>(type: string, load: TLoadCall<T, A, D>, destroy: (link: any) => void, sizeOf?: (link: any) => number, replace: boolean = false) {
    if (createMap.has(type)) {
        if (replace) {
            createMap.get(type).load = load;
            createMap.get(type).destroy = destroy;
            if (sizeOf) {
                createMap.get(type).sizeOf = sizeOf;
            }
        }
        else {
            throw new Error("res_mgr register failed, type has registered, type = " + type);
        }
    }
    createMap.set(type, {
        load,
        destroy,
        sizeOf
    });
    typeGroupList.push(type);

    const resType = createResType(
        [
            {
                min_capacity: 1 * 1024,
                max_capacity: 2 * 1024,
                timeout: 100
            }
        ],
        // (key) => { 
        //     const t = key.split(':')[0];
        //     return typeGroupList.indexOf(t);
        //     // return 1;
        // }
    );

    resTypeMap.set(
        type, 
        resType
    );

    RES_MGR.register_res(resType);
};

/**
 * 类型表
 * @hidden
 */
const createMap: Map<string, { load: (...args: any[]) => Promise<any>, destroy: (link: any) => void, sizeOf: (link: any) => number }> = new Map();

/**
 * 等待加载表
 * @hidden
 */
const waitMap: Map<string, Promise<_Res<ResArg, ResData>>> = new Map();

/**
 * 资源类型， 唯一标识表
 * @hidden
 */
const resTypeMap: Map<string, ResType> = new Map();

/**
 * 记录通过注册的类型
 * @hidden
 */
const typeGroupList: string[] = [];

export const Res = _Res;