
declare var require: any;
declare var module: any;

/**
 ## 资源管理器, 封装了rust版本的资源管理器（pi_lib/res_mgr）

### 需求
* 项目相同资源需要共用，需要一个资源管理器存放资源，并可以同个一个唯一key取出该资源，使资源得到重用
* 缓存一定量的资源，提升运行性能；但缓存又不能过量，否则可能造成程序崩溃或被炒作系统强制杀掉

### pi_lib/res_mgr
pi_lib/res_mgr是一个rust库，它具有以下功能：
* 注册资源类型
* 存放资源注册过的资源类型的实例
* 取出资源
* 设置容量上限；未使用的资源总容量不会超过该容量上限
* 整理未被使用的资源，整理策略：1.删除超时资源；2.容量不足时，优先清除先创建的资源
* 注册资源类型时，同种类型的资源，又可被分为多个子类，并未每个子类型配置最大最小容量
  + 最大容量作用： 
  + 最小容量作用：



/// example
```
// export class TextureRes implements Res {
// 	// 创建纹理资源类型
// 	static TYPE = createResType(
// 		[
// 			{min_capacity: 2 * 1024 * 1024, max_capacity: 16 * 1024 * 1024, timeout: 10*60*1000},
// 			{min_capacity: 2 * 1024 * 1024, max_capacity: 8 * 1024 * 1024, timeout: 10*60*1000}
// 		],
// 		(key: string) => {
// 			if(key.startsWith("common/")){
// 				return 0;
// 			} else {
// 				return 1;
// 			}
// 		});
// 	destroy() {
// 		console.log("纹理资源被销毁");
// 	}
// }

// export class FontRes implements Res {
// 	// 创建字体资源类型
// 	static TYPE = createResType(
// 		[
// 			{min_capacity: 1 * 1024 * 1024, max_capacity: 6 * 1024 * 1024, timeout: 10*60*1000}
// 		],
// 		(_key: string) => {
// 			return 0;
// 		});
// 	destroy() {
// 		console.log("字体资源被销毁");
// 	}
// }


// // 创建一个资源管理器，容量为32M，整理时间间隔为2秒
// export let globalResmgr = new ResMgr(32*1024*1024, 2000);
// // 注册纹理资源类型
// globalResmgr.register_res(TextureRes.TYPE);
// // 注册字体资源类型
// globalResmgr.register_res(FontRes.TYPE);


// ///使用资源管理器
// // 创建资源、获取资源
// var common_texture1 = globalResmgr.create_res(TextureRes.TYPE, "common/a.png", new TextureRes(), 5000);
// var common_texture1_copy = globalResmgr.get_res(TextureRes.TYPE, "common/a.png");

// var uncommon_texture1 = globalResmgr.create_res(FontRes.TYPE, "uncommon/a.png", new FontRes(), 4000);


// // 当资源不使用时，需要释放资源(并不是真的释放，可能被资源管理器缓存起来)
// common_texture1.free();
// common_texture1_copy.free();
// uncommon_texture1.free();
```
*/
import {ResMgr as ResMgr1, Res, ResRef} from "./res_mgr_bg.wasm";
import { now } from "pi_sys/util/time";

export * from "./res_mgr_bg.wasm";

/**
 * 资源管理器
 */
export class ResMgr {
	mgr: ResMgr1;
	isLoop: boolean;
	loop_interval: number;
	
	constructor(total_capacity:number, loop_interval?: number) {
		this.mgr = new ResMgr1(total_capacity);
		this.loop_interval = loop_interval || 1000;
		this.loop_collect();
	}

	/**
	 * 注册某个类型的资源，在资源管理器中存放资源前，必须先为管理器注册该资源，资管管理器才能管理该资源
	 * @param resType 资源类型（构造函数）
	 * @param group_i 该资源类型的分组
	 * @param min_capacity 最小容量， 单位：字节，
	 * @param max_capacity 最大容量，单位：字节
	 * @param time_out 超时时间 单位：ms
	 */
	register_res(resType: ResType): void {
		for(let i = 0; i < resType.all_group.length; i++) {
			let cfg = resType.all_group[i];
			this.mgr.register_res(cal_id(resType.typeId, i), cfg.min_capacity, cfg.max_capacity, cfg.timeout);
		}
	}

	/**
	 * 创建资源
	 * @param resType 已经使用register_res方法注册过的资源类型
	 * @param key 键： 该类型资源的唯一标识， 不同类型的资源，key可以重复
	 * @param res 资源实例
	 * @param cost 资源占用内存（占用容量， 单位：字节）
	 */
	create_res(resType: ResType, key: string, res: Res, cost: number): ResRef {
		let gi = resType.groups(key);
		if (!resType.all_group[gi]) {
			throw new Error("group is not exist, resType: " + JSON.stringify(resType) + ", group: " + gi);
		}
		return this.mgr.create_res(cal_id(resType.typeId, gi), getKey(key), res, cost);
	}

	/**
	 * 获取资源
	 * 该接口获取到的资源，需要通过ResRef的free方法才能释放
	 * @param resType 资源类型
	 * @param key 键： 该类型资源的唯一标识
	 */
	get_res(resType: ResType, key: string): ResRef | undefined {
		let key_num = getKey(key);
		let gi = resType.groups(key);
		if (!resType.all_group[gi]) {
			throw new Error("group is not exist, resType: " + JSON.stringify(resType) + ", group: " + gi);
		}
		return this.mgr.get_res(cal_id(resType.typeId, gi), key_num);
	}

	/**
	 * 循环整理资源，资源管理器实例化时，会开启整理循环， 外部可通过设置资源管理器的loop_interval资源来控制整理间隔时间
	 */
	private loop_collect = (): void => {
		this.mgr.collect(now());
		setTimeout(this.loop_collect, this.loop_interval);
	};
}

/**
 * 资源类型， 唯一标识某类资源
 */
export interface ResType {
	/**
	 * 对每种资源类型，该值是唯一的，用于区分不同的资源类型（由于资源管理器是rust， 使用数字与rust交互）
	 */
	typeId: number,
	/**
	 * 分组的配置，用于将资源分为多组，每组可配置单独的最大最小容量
	 * 比如： 纹理资源类型，可以按照常用ui图标、不常用ui图片、常用场景图片、不常用场景图片等分为多组，
	 * ------分别为它们配置不同的最大最小容量和超时时间，通常将常用图片配置的容量更大，超时时间更长，
	 * ------使得常用纹理资源相交于不常用纹理资源，在资源管理器容量不足时，更不容易被释放。
	 */
	all_group: ResCfg[],

	/**
	 * 分组, 返回all_groups的某个索引
	 * 对于某个资源被分到什么组，通常可以通过key来决定。
	 * 比如： 假如我们在all_group中配置了常用ui图标和不常用ui图片两个组，groups方法可以实现为： 
	 * 	```
	 * 	// 这里假定_key是纹理对应图片的路径，常用图片被放置在名为“common”的跟目录中
	 * 	if(_key.startWith("common/")){
	 * 		return 0;
	 * 	} else {
	 * 		return 1
	 * 	}
	 * 	```
	 * @param _key 键： 资源的唯一标识
	 */
	groups(_key: string): number;

}

/**
 * 资源分组配置
 * 每种资源类型，都需要有该配置
 */
export interface ResCfg {
	/**
	 * 最小容量，资源管理器为其保留的最小容量，当资源管理器容量不足时，即便该类资源未占用完该容量，
	 * 也不会将剩余空间让出给其它类型的资源
	 * 如： 资源管理器总容量为10， A类和B类资源配置的最小容量都为3，当前，A类资源占容量为1，B类资源占用容量为7，
	 * -----此时，若一个大小为1的B类资源不再被任何地方使用，资源管理器会尽量将其缓存起来；资源管理器当前被占用的总容量为8， 实际是可以缓存该资源的，
	 * -----但，由于A资源配置最小容量为3，A资源自身占用1，剩余的大小为2的容量，不会让出给B使用，
	 * -----因此，该资源会被直接释放掉，不会被缓存。
	 * 配置该字段得好处在于，一些资源类型加载时机比较晚，但该资源在程序后续运行中的缓存十分重要，当其被加载缓存时，如果资源管理器已经为其预留了一部分空间，
	 * 则不会再调整其它资源的缓存空间。（注： 似乎只是在少数时刻减少了资源管理器的调整，节省一些性能开销，但是否必要？）
	 */
	min_capacity: number,

	/**
	 * 最大容量。主要作为和其它资源类型争夺资源管理器容量的权重，并不表示该类型的资源能缓存的最大容量
	 * 如： 纹理资源配置的最大容量为10，材质资源配置的最大容量为2，当资源管理器容量不足时，纹理资源能争夺到更多的容量。
	 */
	max_capacity: number,
	/**
	 * 超时时间
	 * 资源管理器会定期扫描所有缓存资源，无论当前容量是否紧缺，都会释放掉已经超时的资源。
	 * 该字段用于指明某一类资源的超时时间
	 */
	timeout: number,
}

/**
 * 提供一个便携的方法，用于创建一个资源类型
 * @param all_group 对该资源的分组描述，该数组长度至少为1，否则发生异常
 * @param groups 可选；以资源的key为参数，返回all_group的一个索引，来表明该资源应该被归为哪一组，
 * --------------如果不传该字段，则为其使用一个默认方法，该方法返回0（因此，all_group的长度至少为1）
 */
export const createResType = (all_group: ResCfg[], groups?: (key: string) => number): ResType => {
	return {
		typeId: ++type_id, // 因为typeId需要唯一，因此这里以一个递增的数字来表示
		all_group: all_group,
		groups: groups || defaultGroups, 
	}
}

/**
 * 默认的分组方法， 仅将资源分为1组，分组索引为0
 */
const defaultGroups = (_key: string) => {
	return 0;
}


/******************************************* 内部方法 *********************************************/
/**
 * 根据typeId和groupId计算资源的唯一id
 */
const cal_id = (typeId: number, group_i: number) => {
	return typeId << 16 + group_i;
}

/**
 * 为key映射一个数组，并返回该数字（js于rust的字符串交互比较慢，因此映射为数字）
 */
const getKey = (key: string) => {
	let r = keyMap.get(key);
	if(!r) {
		r = ++KEY_NUM;
		keyMap.set(key, r);
	}
	return r;
}

// 将key映射为数字（性能考虑底层只接受数字）
const keyMap: Map<string, number> = new Map();
let type_id = 0;
let KEY_NUM = 0;


