/**
 * 声音播放模块
 * ```
 * 注意，时间单位是秒, 可以使用小数表示毫秒
 * 注意划分了 背景音乐 与 音效 两种播放内容
 * ```
 * @example
 * ```
 * import { BGMusic, SoundResMgr } from 'pi_utils/res/sound'
 * 
 * // 播放背景音乐
 * BGMusic.play('a.mp3');
 * BGMusic.pause('a.mp3');
 * BGMusic.stop();
 * // 播放背景音乐
 * BGMusic.play('b.mp3');
 * // 静音
 * BGMusic.setMuted(true);
 * // 取消静音
 * BGMusic.setMuted(false);
 * 
 * // 播放音效
 * SoundResMgr.play('1.mp3', 1.0);
 * ```
 * @packageDocumentation
 */

// ============================== 导入
import { ResTab } from './res_mgr';
import { PISYS } from 'pi_sys/index';
import { PiAudio } from 'pi_sys/media/audio';
import { RES_AUDIO } from './res_load';

/**
 * @description 声音状态类型
 */
export enum StateType {
    /**
     * 初始化
     */
    INIT = 0, // 
    /**
     * 就绪
     */
    READY, // 
    /**
     * 播放
     */
    PLAY, // 
    /**
     * 暂停
     */
    PAUSE // 
}

/**
 * 步进模式
 */
export type StepType = 'linear' | 'quadIn' | 'quadOut' | 'quadInOut';

/**
 * 背景音乐
 * ```
 * 使用audio标签播放，
 * 只存在一个，
 * 每次切换音乐场景时，在点击事件中触发
 * ios 的交互操作不能放在回调中，故而只能采用使用相对资源路径
 * ```
 */
export namespace BGMusic {
    /**
     * 当前播放音乐的路径
     * @hidden
     */
    let currpath: string = null;

    /**
     * 音乐节点，不用重复创建
     * @hidden
     */
    export const audioMap: Map<string, PISYS.PiAudio> = new Map();

    /**
     * 资源服务器cdn（实例化时设置）
     * @hidden
     */
    export let server: string = "null";

    /**
     * 静音，用于设置
     * @hidden
     */
    let _muted: boolean = false;
    // /**
    //  *
    //  * @param s 资源服务器 PI.server
    //  */
    // constructor(s: string) {
    //     server = s;
    // }

    /**
     * 停止当前音乐，播放目标音乐
     * @param path 资源路径
     * @param volume 音量
     * @param loop 循环
     * @param stepTime 渐进时间
     * @param stepType 渐进类型
     */
    export function play(path: string, volume: number = 1, loop: boolean = true, stepTime: number = 0, stepType?: StepType) {
        if (!server) {
            throw new Error('set server first!');
        }
        if (_muted) {
            console.warn('muted set');

            return;
        }
        if (path === currpath) {
            console.warn(`${path} is playing`);

            return;
        }

        // 停止当前资源
        stop();
        currpath = path;
        let a = audioMap.get(path);

        if (a) {
            a.playAsBGM(volume, loop, stepTime, stepType);
        } else {
            SoundResMgr.resTab.load(RES_AUDIO, path, []).then(res => {
                const audio: PISYS.PiAudio = res.link;
                audioMap.set(path, audio);
                if (currpath == path) {
                    audio.playAsBGM(volume, loop, stepTime, stepType);
                }
            });
        }
    }

    /**
     * 停止
     */
    export function stop() {
        if (!currpath) {
            return;
        }

        const now = audioMap.get(currpath);
        if (!now) {
            console.warn(`${currpath} not found`);

            return;
        }

        now.stopAsBGM();
        currpath = null;
    }

    /**
     * 暂停目标音频
     * @param path 目标音频路径
     */
    export function pause(path: string) {
        const a = audioMap.get(path);
        if (!a) {
            console.warn(`${path} not found`);

            return;
        }
        a.pauseAsBGM();

    }

    /**
     * 切换静音
     * @param muted 是否静音
     */
    export function setMuted(muted: boolean) {
        _muted = muted;
        PiAudio.enable(!muted);
    }

}

/**
 * 音效资源管理对象 - 播放音效
 * ```
 * 只提供播放接口
 * 可同时调用多次播放 - 某些平台限制10个
 * ```
 */
export namespace SoundResMgr {
    /**
     * 资源表
     */
    export const resTab = new ResTab();

    /**
     * 音量， 0-1之间
     * @hidden 弃用
     */
    export let volume: number = 1;

    /**
     * 
     * @param src 目标音频路径
     * @param volume 音量
     * @param delay 延时
     * @param repeat 重复次数 - 某些运行环境无效
     * @param repeatDelay 重复时延时 - 某些运行环境无效
     */
    export function play(src: string, volume: number, delay?: number, repeat?: number, repeatDelay?: number): void {

        resTab.load(RES_AUDIO, src, []).then((res) => {
            const audio: PISYS.PiAudio = res.link;
            audio.volume = volume;
            audio.play(delay, repeat, repeatDelay);
            audio.onended = () => {
                if (audio.arr.length == 0) {
                    resTab.delete(res);
                }
            }
        });

    }
}

// ============================== 本地
