/**
 * 实现了 *.cfg.json 、 *.jcfg 类型文件的 加载时处理过程
 * @example 实现示例
 * ```
 * const XXX_SUFFIX = "xxx";
 * function xxxHandler(): Promise<any> {
 *     // 注册 XXX_SUFFIX 类型文件处理钩子
 *     return PISYS.setCfgHandler(
 *         XXX_SUFFIX,
 *         (_FILE: string, data: Uint8Array): Promise<any> => {
 *             return new Promise((resolve, reject) => {
 *                 // 具体处理数据的逻辑 - 例如解析为 json 保存到自定义的 Map
 *                 // (必须)处理结束 调用 resolve() - 数据自定，可以传null
 *                 // (必须)处理重新错误 调用 reject() - 自定错误信息
 *             });
 *         }
 *     );
 * } 
 * ```
 * @packageDocumentation
 */

import { PISYS } from "pi_sys/index";
import { utf8Decode } from "./util";
import { parseCfg, parseJCfg } from "./cfg";

/**
 * @hidden
 */
declare var _$pi;

/**
 * 整体注册
 * @hidden pi_uitls/index 调用
 */
export const registerCfgHandler = () => {
    sCfgHandle();
    jCfgHandle();
}

/**
 * @hidden
 */
const SCFG_SUFFIX = "cfg.json";
/**
 * @hidden
 */
const sCfgHandle = (): Promise<any> => {
    return PISYS.setCfgHandler(SCFG_SUFFIX, (_file: string, data: Uint8Array): Promise<any> => {
        return new Promise((resolve, reject) => {
            _$pi.exec(() => {
                try {
                    const cfgs = JSON.parse(utf8Decode(data));
                    _$pi.exec(() => {
                        parseCfg(cfgs).then((r) => {
                            resolve(r);
                        }).catch((e) => {
                            reject(e);
                        });
                    });
                } catch (e) {
                    reject(e);
                }
            });
        });
    });
}

/**
 * @hidden
 */
const JCFG_SUFFIX = "jcfg";
/**
 * @hidden
 */
const jCfgHandle = (): Promise<any> => {
    return PISYS.setCfgHandler(JCFG_SUFFIX, (_file: string, data: Uint8Array): Promise<any> => {
        return new Promise((resolve, reject) => {
            _$pi.exec(() => {
                try {
                    const cfgs = JSON.parse(utf8Decode(data));
                    _$pi.exec(() => {
                        parseJCfg(cfgs).then((r) => {
                            resolve(r);
                        }).catch((e) => {
                            reject(e);
                        });
                    });
                } catch (e) {
                    reject(e);
                }
            });
        });
    });
}

