/**
 * 事件广播模块
 * 
 * @packageDocumentation
 */

// ============================== 导入
import { now as timeNow } from '../lang/time';
import { Func } from '../lang/type';
// import { PISYS } from 'pi_sys/index';
import { arrInsert, arrRemove, call, objCall } from './util';

// ============================== 导出

/**
 * `timeout` 毫秒以上的事件会打印
 */
const timeout = 3;

/**
 * @description 处理器返回值枚举
 */
export enum HandlerResult {
    /**
     * ok
     */
    OK = 0,
    /**
     * 从处理器列表删除自身，以后不会收到事件
     */
    REMOVE_SELF, // 
    /**
     * 结束此次事件调用，不继续调用处理器
     */
    BREAK_OK, // 
    /**
     * 结束此次事件调用，不继续调用处理器，从处理器列表删除自身，以后不会收到事件
     */
    BREAK_REMOVE_SELF // 
}

/**
 * 定义 事件调用 的实现细节
 */
export type Handler = (e: any) => HandlerResult;

/**
 * 定义 事件调用列表 数据结构
 */
export interface HandlerList extends Handler, HandlerArray {
    // (e: any): HandlerResult;
    size(): number;
    add(h: Handler): void;
    remove(h: Handler): boolean;
    clear(): void;
    __proto__: HandlerArray;
}

/**
 * 创建事件处理器列表
 * @example
 */
export const createHandlerList = (): Handler => {
    const list: Handler = (args): HandlerResult => {
        let i;
        let handler: Handler;
        let r;
        let delIndex = -1;
        const arr = (<HandlerList>list).array;
        const n = arr.length;
        (<HandlerList>list).handling++;
        for (i = n - 1; i >= 0; i--) {
            handler = arr[i];
            if (handler) {
                r = call1(handler, args);
                if (!r) {
                    continue;
                } else if (r === HandlerResult.REMOVE_SELF) {
                    arr[i] = null;
                    delIndex = i;
                } else if (r === HandlerResult.BREAK_OK) {
                    break;
                } else if (r === HandlerResult.BREAK_REMOVE_SELF) {
                    arr[i] = null;
                    delIndex = i;
                    break;
                }
            } else {
                delIndex = i;
            }
        }
        (<HandlerList>list).handling--;
        if (delIndex >= 0 && !(<HandlerList>list).handling) {
            for (i = delIndex + 1; i < n; ++i) {
                handler = arr[i];
                if (handler) {
                    arr[delIndex] = handler;
                    ++delIndex;
                }
            }
            (<HandlerList>list).count = delIndex;
            arr.length = delIndex;
        }

        return r || HandlerResult.BREAK_OK;
    };
    (<HandlerList>list).handling = 0;
    (<HandlerList>list).count = 0;
    (<HandlerList>list).array = [];
    (<HandlerList>list).__proto__ = HandlerArray.prototype;

    return list;
};

/**
 * 创建事件处理器表
 * @example
 */
export class HandlerMap {

    /**
     * 事件处理器映射表
     * @hidden
     */
    private map: Map<string, HandlerList> = new Map();
    /**
     * 事件通知
     * @param _type 事件类型
     * @param args 响应数据
     * @returns 响应是否成功结束
     */
    public notify(_type: string, args: any): boolean {
        const list = this.map.get(_type);
        if (!list) {
            return false;
        }
        // 这是一个bug,如果传入的参数本来就是数组，就会解析出错
        Array.isArray(args) ? list([args]) : list(args);

        return true;
    }
    /**
     * 获得事件处理器表的长度
     * @param _type 事件类型
     * @return 长度
     */
    public size(_type: string): number {
        let list;
        let n = 0;
        for (list of this.map.values()) {
            n += list.size();
        }

        return n;
    }
    /**
     * 添加事件处理器
     * @param _type 事件类型
     * @param h 处理函数
     */
    public add(_type: string, h: Handler): void {
        let list: any;
        const map = this.map;
        if (!(h && _type)) {
            return;
        }
        list = map.get(_type);
        if (!list) {
            list = createHandlerList();
            map.set(_type, list);
        }
        list.add(h);
    }
    /**
     * 删除事件处理器
     * @param _type 事件类型
     * @param h 处理函数
     * @returns 是否成功
     */
    public remove(_type: string, h: Handler): boolean {
        let list;
        const map = this.map;
        if (!h) {
            if (!_type) {
                return false;
            }

            return map.delete(_type);
        }
        if (!_type) {
            for (list of this.map.values()) {
                if (!list.remove(h)) {
                    continue;
                }
                if (list.size() === 0) {
                    map.delete(_type);

                    return true;
                }
            }

            return false;
        }
        list = map.get(_type);
        if (!list) {
            return false;
        }
        if (!list.remove(h)) {
            return false;
        }
        if (list.size() === 0) {
            map.delete(_type);
        }

        return true;
    }
    /**
     * 删除事件处理器
     */
    public clear(): void {
        this.map.clear();
    }
}

/**
 * 事件处理器表
 * @example
 */
export class HandlerTable {
    /**
     * 事件处理器表
     * @tip 必须要赋初值，不然new出来的实例里面是没有这些属性的
     * @hidden
     */
    private handlerMap: HandlerMap = null; // 
    /**
     * 通知组件上的事件监听器
     * @param eventType 事件类型
     * @param args 响应数据 - 一个对象,非数组
     * @returns 响应是否成功结束
     */
    public notify(eventType: string, args: any): boolean {
        if (this[eventType]) {
            return objCall1(this, eventType, args);
        }
        const map = this.handlerMap;
        if (!map) {
            return;
        }
        const r = map.notify(eventType, args);
        if (r) {
            return r;
        }

        return map.notify('*', args);
    }
    /**
     * 添加事件处理器
     * @param _type 事件类型
     * @param h 处理函数
     */
    public addHandler(_type: string, h: Handler): void {
        let map = this.handlerMap;
        if (!map) {
            map = this.handlerMap = new HandlerMap();
        }
        map.add(_type, h);
    }
    /**
     * 删除事件处理器
     * @param _type 事件类型
     * @param h 处理函数
     * @returns 是否成功
     */
    public removeHandler(_type: string, h: Handler): boolean {
        return this.handlerMap && this.handlerMap.remove(_type, h);
    }
    /**
     * 清空事件处理器
     */
    public clearHandler(): void {
        return this.handlerMap && this.handlerMap.clear();
    }
}

/**
 * 事件监听器表
 * @example
 */
export class ListenerList<T> {
    /**
     * @hidden
     */
    private list: Func<T>[] = [];

    /**
     * @description 通知列表上的每个事件监听器
     * @param arg 事件响应的数据
     * @example
     */
    public notify(arg: T) {
        const r = this.list;
        for (const f of r) {
            f(arg);
        }
    }
    /**
     * 获取事件监听器列表的长度
     * @returns 监听器列表的长度
     */
    public size(): number {
        return this.list.length;
    }
    /**
     * 添加事件监听器
     * @param f 事件监听函数
     */
    public add(f: Func<T>): void {
        this.list = arrInsert(this.list, f);
    }
    /**
     * 删除事件监听器
     * @param f 事件监听函数
     * @returns 删除是否成功
     */
    public remove(f: Func<T>): boolean {
        const old = this.list;
        const r = arrRemove(this.list, f);
        const b = r === old;
        this.list = r;

        return b;
    }
    /**
     * 清空事件监听器列表
     */
    public clear(): void {
        this.list = [];
    }
}

// ============================== 本地

/**
 * 函数调用
 * @param func 函数
 * @param args 参数数组
 * @returns 函数执行返回值
 * @hidden
 */
const call1 = (func: Function, args: any[]) => {
    let r;
    const start = timeNow();
    try {
        r = call(func, args);
    } catch (ex) {
        // return PISYSPISYS.Log.cc.warn() && PISYS.Log.log('event, ex: ', ex, ', func: ', func, args);
    }
    const end = timeNow();
    if (end - start > timeout) {
        // PISYS.Log.cc.debug() && PISYS.Log.log(`event slow, cost: ${end - start}`, func, args);
    }

    return r;
};

/**
 */
/**
 * 对象方法调用
 * @param {any} obj 对象
 * @param {string} func 方法名
 * @param {any[]} args 参数数组
 * @returns {any} 函数执行返回值
 * @hidden
 */
const objCall1 = (obj: any, func: string, args: any[]) => {
    let r;
    const start = timeNow();
    try {
        r = objCall(obj, func, args);
    } catch (ex) {
        // return PISYS.Log.cc.warn() && PISYS.Log.log('event, ex: ', ex, ', func: ', obj, func, args);
    }
    const end = timeNow();
    if (end - start > timeout) {
        // PISYS.Log.cc.debug() && PISYS.Log.log(`event slow, cost: ${end - start}`, obj, func, args);
    }

    return r;
};

/**
 * @TODO 以后改成树结构-并且是写时复制的，就可以任意重入。而且删除效率高。js对象不能直接比较大小，可以转成字符串后的hash来比较大小。如果是乱序执行，则只需要1个树。如果是按照放入的顺序执行，则需要2个树。sbtree或fingertree
 * @TODO 改为用set实现
 */
class HandlerArray {
    /**
     * @hidden
     */
    public handling: number = 0;
    /**
     * @hidden
     */
    public count: number = 0;
    /**
     * @hidden
     */
    public array: Handler[] = [];
    /**
     * 获得事件处理器列表的长度
     * @returns {number} 事件处理器列表的长度
     */
    public size(): number {
        return this.count;
    }
    /**
     * 添加事件处理器
     * @param {Handler} 事件处理器
     */
    public add(handler: Handler) {
        this.array.push(handler);
        this.count += 1;
    }
    /**
     * 删除事件处理器
     * @param {Handler} 事件处理器
     * @returns {boolean} 删除是否成功
     */
    public remove(handler: Handler): boolean {
        let i;
        const arr = this.array;
        for (i = arr.length - 1; i >= 0; --i) {
            if (arr[i] === handler) {
                arr[i] = null;
                this.count -= 1;

                return true;
            }
        }

        return false;
    }
    /**
     * 清空事件处理器
     */
    public clear() {
        this.array = [];
        this.count = 0;
    }
}
