/**
 Starting with version 2.0, this file "boots" Jasmine, performing all of the necessary initialization before executing the loaded environment and all of a project's specs. This file should be loaded after `jasmine.js` and `jasmine_html.js`, but before any project source files or spec files are loaded. Thus this file can also be used to customize Jasmine for a project.

 If a project is using Jasmine via the standalone distribution, this file can be customized directly. If a project is using Jasmine via the [Ruby gem][jasmine-gem], this file can be copied into the support directory via `jasmine copy_boot_js`. Other environments (e.g., Python) will have different mechanisms.

 The location of `boot.js` can be specified and/or overridden in `jasmine.yml`.

 [jasmine-gem]: http://github.com/pivotal/jasmine-gem
 */

(function () {
  var jasmineRequire = require('./jasmine');

  /**
   * ## Require &amp; Instantiate
   *
   * Require Jasmine's core files. Specifically, this requires and attaches all of Jasmine's code to the `jasmine` reference.
   */
  var jasmine = jasmineRequire.core(jasmineRequire),
    global = jasmine.getGlobal();
  global.jasmine = jasmine;
  var fn = global.setTimeout;
  /**
   * 这里需要重载,否则当arg2为undefined,底层会出错
   * @param {*} arg1 
   * @param {*} arg2 
   */
  global.setTimeout = function (arg1, arg2) {
    if (arg2 === undefined) arg2 = 0;
    fn(arg1, arg2)
  }

  /**
   * Since this is being run in a browser and the results should populate to an HTML page, require the HTML-specific Jasmine code, injecting the same reference.
   */
  // jasmineRequire.html(jasmine);

  /**
   * Create the Jasmine environment. This is used to run all specs in a project.
   */
  var env = jasmine.getEnv();

  /**
   * ## The Global Interface
   *
   * Build up the functions that will be exposed as the Jasmine public interface. A project can customize, rename or alias any of these functions as desired, provided the implementation remains unchanged.
   */
  var jasmineInterface = jasmineRequire.interface(jasmine, env);

  /**
   * Add all of the Jasmine global/public interface to the global scope, so a project can use the public interface directly. For example, calling `describe` in specs instead of `jasmine.getEnv().describe`.
   */
  extend(global, jasmineInterface);

  /**
   * The `jsApiReporter` also receives spec results, and is used by any environment that needs to extract the results  from JavaScript.
   */
  var myReporter = {
    totalSpecsDefined: 0,  // 单元测试总数
    failedResults: [],
    jasmineStarted: function (suiteInfo) {
      console.log("================= jasmineStarted ===================");
      this.totalSpecsDefined = suiteInfo.totalSpecsDefined;
    },

    suiteStarted: function (result) {
    },

    specStarted: function (result) {
    },

    specDone: function (result) {
      if (result.status === 'failed') {
        this.failureCount++;
        this.failedResults.push(result);
      }
    },

    suiteDone: function (result) {

    },

    jasmineDone: function (result) {
      for (var k = 0; k < this.failedResults.length; k++) {
        var result = this.failedResults[k];
        // console.log("\"" + result.description + "\"" + ' was ' + result.status);
        for (var i = 0; i < result.failedExpectations.length; i++) {
          console.log('Failure: ');
          console.log((k + 1) + " ) " + result.fullName);
          console.log("Message:");
          console.log("     " + result.failedExpectations[i].message);
          console.log("Stack:");
          console.log("     " + result.failedExpectations[i].stack);
          // console.log('Failure: ' + result.failedExpectations[i].message);
          // console.log(result.failedExpectations[i].stack);
        }
      }
      console.log("Total spec " + this.totalSpecsDefined + " , " + this.failedResults.length + " failed");
      console.log("================= jasmineDone ===================");
    }
  };

  env.addReporter(jasmineInterface.jsApiReporter);
  env.addReporter(myReporter);

  /**
   * Helper function for readability above.
   */
  function extend(destination, source) {
    for (var property in source) destination[property] = source[property];
    return destination;
  }

  /**
   * 说明浏览器环境
   */
  if (jasmineRequire.html) {
    jasmineRequire.html(jasmine);

    var queryString = new jasmine.QueryString({
      getWindowLocation: function () { return window.location; }
    });

    var filterSpecs = !!queryString.getParam("spec");

    var config = {
      failFast: queryString.getParam("failFast"),
      oneFailurePerSpec: queryString.getParam("oneFailurePerSpec"),
      hideDisabled: queryString.getParam("hideDisabled")
    };

    var random = queryString.getParam("random");

    if (random !== undefined && random !== "") {
      config.random = random;
    }

    var seed = queryString.getParam("seed");
    if (seed) {
      config.seed = seed;
    }

    /**
     * ## Reporters
     * The `HtmlReporter` builds all of the HTML UI for the runner page. This reporter paints the dots, stars, and x's for specs, as well as all spec names and all failures (if any).
     */
    var htmlReporter = new jasmine.HtmlReporter({
      env: env,
      navigateWithNewParam: function (key, value) { return queryString.navigateWithNewParam(key, value); },
      addToExistingQueryString: function (key, value) { return queryString.fullStringWithNewParam(key, value); },
      getContainer: function () { return document.body; },
      createElement: function () { return document.createElement.apply(document, arguments); },
      createTextNode: function () { return document.createTextNode.apply(document, arguments); },
      timer: new jasmine.Timer(),
      filterSpecs: filterSpecs
    });


    env.addReporter(htmlReporter);

    /**
     * Filter which specs will be run by matching the start of the full name against the `spec` query param.
     */
    var specFilter = new jasmine.HtmlSpecFilter({
      filterString: function () { return queryString.getParam("spec"); }
    });

    config.specFilter = function (spec) {
      return specFilter.matches(spec.getFullName());
    };

    env.configure(config);

    /**
     * Setting up timing functions to be able to be overridden. Certain browsers (Safari, IE 8, phantomjs) require this hack.
     */
    window.setTimeout = window.setTimeout;
    window.setInterval = window.setInterval;
    window.clearTimeout = window.clearTimeout;
    window.clearInterval = window.clearInterval;

    /**
     * ## Execution
     *
     * Replace the browser window's `onload`, ensure it's called, and then run all of the loaded specs. This includes initializing the `HtmlReporter` instance and then executing the loaded Jasmine environment. All of this will happen after all of the specs are loaded.
     */
    htmlReporter.initialize();
  }

  exports.jasmineExecute = env.execute
}());
